@@ -16,20 +16,26 @@
 # under the License.
 import collections
 import datetime
+import itertools
 import operator
+import struct
 import uuid
 
 from concurrent import futures
 import iso8601
+import msgpack
 from oslo_config import cfg
 from oslo_log import log
-from oslo_serialization import msgpackutils
 from oslo_utils import timeutils
+import pandas
 import six
+import six.moves
 from tooz import coordination
 
 from gnocchi import carbonara
 from gnocchi import storage
+from gnocchi import utils
+
 
 OPTS = [
     cfg.IntOpt('aggregation_workers_number',
@@ -46,19 +52,37 @@ OPTS = [
 LOG = log.getLogger(__name__)
 
 
+class CorruptionError(ValueError):
+    """Data corrupted, damn it."""
+
+    def __init__(self, message):
+        super(CorruptionError, self).__init__(message)
+
+
 class CarbonaraBasedStorage(storage.StorageDriver):
     MEASURE_PREFIX = "measure"
+    UPGRADE_BATCH_SIZE = 1000
 
     def __init__(self, conf):
         super(CarbonaraBasedStorage, self).__init__(conf)
+        self.coord = coordination.get_coordinator(
+            conf.coordination_url,
+            str(uuid.uuid4()).encode('ascii'))
+        self.aggregation_workers_number = conf.aggregation_workers_number
+        if self.aggregation_workers_number == 1:
+            # NOTE(jd) Avoid using futures at all if we don't want any threads.
+            self._map_in_thread = self._map_no_thread
+        else:
+            self._map_in_thread = self._map_in_futures_threads
+        self.start()
+
+    @utils.retry
+    def start(self):
         try:
-            self.coord = coordination.get_coordinator(
-                conf.coordination_url,
-                str(uuid.uuid4()).encode('ascii'))
             self.coord.start(start_heart=True)
         except Exception as e:
-            raise storage.StorageError("Unable to start coordinator: %s" % e)
-        self.aggregation_workers_number = conf.aggregation_workers_number
+            LOG.error("Unable to start coordinator: %s" % e)
+            raise utils.Retry(e)
 
     def stop(self):
         self.coord.stop()
@@ -68,26 +92,67 @@ class CarbonaraBasedStorage(storage.StorageDriver):
         return self.coord.get_lock(lock_name)
 
     @staticmethod
-    def _get_measures(metric, timestamp_key, aggregation, granularity):
+    def _get_measures(metric, timestamp_key, aggregation, granularity,
+                      version=3):
         raise NotImplementedError
 
     @staticmethod
-    def _get_unaggregated_timeserie(metric):
+    def _get_unaggregated_timeserie(metric, version=3):
+        raise NotImplementedError
+
+    def _get_unaggregated_timeserie_and_unserialize(
+            self, metric, block_size, back_window):
+        """Retrieve unaggregated timeserie for a metric and unserialize it.
+
+        Returns a gnocchi.carbonara.BoundTimeSerie object. If the data cannot
+        be retrieved, returns None.
+
+        """
+        with timeutils.StopWatch() as sw:
+            raw_measures = (
+                self._get_unaggregated_timeserie(
+                    metric)
+            )
+            LOG.debug(
+                "Retrieve unaggregated measures "
+                "for %s in %.2fs"
+                % (metric.id, sw.elapsed()))
+        try:
+            return carbonara.BoundTimeSerie.unserialize(
+                raw_measures, block_size, back_window)
+        except ValueError:
+            raise CorruptionError(
+                "Data corruption detected for %s "
+                "unaggregated timeserie" % metric.id)
+
+    @staticmethod
+    def _store_unaggregated_timeserie(metric, data, version=3):
         raise NotImplementedError
 
     @staticmethod
-    def _store_unaggregated_timeserie(metric, data):
+    def _store_metric_measures(metric, timestamp_key, aggregation,
+                               granularity, data, offset=None, version=3):
         raise NotImplementedError
 
     @staticmethod
-    def _store_metric_measures(metric, timestamp_key,
-                               aggregation, granularity, data):
+    def _delete_unaggregated_timeserie(metric, version=3):
         raise NotImplementedError
 
     @staticmethod
-    def _list_split_keys_for_metric(metric, aggregation, granularity):
+    def _list_split_keys_for_metric(metric, aggregation, granularity,
+                                    version=None):
         raise NotImplementedError
 
+    @staticmethod
+    def _version_check(name, v):
+        """Validate object matches expected version.
+
+        Version should be last attribute and start with 'v'
+        """
+        attrs = name.split("_")
+        return not v or (not attrs[-1].startswith('v') if v == 2
+                         else attrs[-1] == 'v%s' % v)
+
     def get_measures(self, metric, from_timestamp=None, to_timestamp=None,
                      aggregation='mean', granularity=None):
         super(CarbonaraBasedStorage, self).get_measures(
@@ -110,12 +175,13 @@ class CarbonaraBasedStorage(storage.StorageDriver):
                                       aggregation, granularity):
         data = self._get_measures(metric, key, aggregation, granularity)
         try:
-            return carbonara.TimeSerie.unserialize(data)
+            return carbonara.AggregatedTimeSerie.unserialize(
+                data, key, aggregation, granularity)
         except ValueError:
-                LOG.error("Data corruption detected for %s "
-                          "aggregated `%s' timeserie, granularity `%s' "
-                          "around time `%s', ignoring."
-                          % (metric.id, aggregation, granularity, key))
+            LOG.error("Data corruption detected for %s "
+                      "aggregated `%s' timeserie, granularity `%s' "
+                      "around time `%s', ignoring."
+                      % (metric.id, aggregation, granularity, key))
 
     def _get_measures_timeserie(self, metric,
                                 aggregation, granularity,
@@ -134,39 +200,23 @@ class CarbonaraBasedStorage(storage.StorageDriver):
             all_keys = self._list_split_keys_for_metric(
                 metric, aggregation, granularity)
         except storage.MetricDoesNotExist:
-            # This can happen if it's an old metric with a TimeSerieArchive
-            all_keys = None
-
-        if not all_keys:
-            # It does not mean we have no data: it can be an old metric with a
-            # TimeSerieArchive.
-            try:
-                data = self._get_metric_archive(metric, aggregation)
-            except (storage.MetricDoesNotExist,
-                    storage.AggregationDoesNotExist):
-                # It really does not exist
-                for d in metric.archive_policy.definition:
-                    if d.granularity == granularity:
-                        return carbonara.AggregatedTimeSerie(
-                            sampling=granularity,
-                            aggregation_method=aggregation,
-                            max_size=d.points)
-                raise storage.GranularityDoesNotExist(metric, granularity)
-            else:
-                archive = carbonara.TimeSerieArchive.unserialize(data)
-                # It's an old metric with an TimeSerieArchive!
-                for ts in archive.agg_timeseries:
-                    if ts.sampling == granularity:
-                        return ts
-                raise storage.GranularityDoesNotExist(metric, granularity)
+            for d in metric.archive_policy.definition:
+                if d.granularity == granularity:
+                    return carbonara.AggregatedTimeSerie(
+                        sampling=granularity,
+                        aggregation_method=aggregation,
+                        max_size=d.points)
+            raise storage.GranularityDoesNotExist(metric, granularity)
 
         if from_timestamp:
-            from_timestamp = carbonara.AggregatedTimeSerie.get_split_key(
-                from_timestamp, granularity)
+            from_timestamp = str(
+                carbonara.SplitKey.from_timestamp_and_sampling(
+                    from_timestamp, granularity))
 
         if to_timestamp:
-            to_timestamp = carbonara.AggregatedTimeSerie.get_split_key(
-                to_timestamp, granularity)
+            to_timestamp = str(
+                carbonara.SplitKey.from_timestamp_and_sampling(
+                    to_timestamp, granularity))
 
         timeseries = filter(
             lambda x: x is not None,
@@ -184,27 +234,113 @@ class CarbonaraBasedStorage(storage.StorageDriver):
             timeseries=timeseries,
             max_size=points)
 
-    def _add_measures(self, aggregation, archive_policy_def,
-                      metric, timeserie):
-        ts = self._get_measures_timeserie(metric, aggregation,
-                                          archive_policy_def.granularity,
-                                          timeserie.first, timeserie.last)
-        ts.update(timeserie)
-        for key, split in ts.split():
-            self._store_metric_measures(metric, key, aggregation,
-                                        archive_policy_def.granularity,
-                                        split.serialize())
+    def _store_timeserie_split(self, metric, key, split,
+                               aggregation, archive_policy_def,
+                               oldest_mutable_timestamp):
+        # NOTE(jd) We write the full split only if the driver works that way
+        # (self.WRITE_FULL) or if the oldest_mutable_timestamp is out of range.
+        write_full = self.WRITE_FULL or next(key) < oldest_mutable_timestamp
+        key_as_str = str(key)
+        if write_full:
+            try:
+                existing = self._get_measures_and_unserialize(
+                    metric, key_as_str, aggregation,
+                    archive_policy_def.granularity)
+            except storage.AggregationDoesNotExist:
+                pass
+            else:
+                if existing is not None:
+                    if split is None:
+                        split = existing
+                    else:
+                        split.merge(existing)
 
-        if ts.last and archive_policy_def.timespan:
+        offset, data = split.serialize(key, compressed=write_full)
+
+        return self._store_metric_measures(
+            metric, key_as_str, aggregation, archive_policy_def.granularity,
+            data, offset=offset)
+
+    def _add_measures(self, aggregation, archive_policy_def,
+                      metric, grouped_serie,
+                      previous_oldest_mutable_timestamp,
+                      oldest_mutable_timestamp):
+        ts = carbonara.AggregatedTimeSerie.from_grouped_serie(
+            grouped_serie, archive_policy_def.granularity,
+            aggregation, max_size=archive_policy_def.points)
+
+        # Don't do anything if the timeserie is empty
+        if not ts:
+            return
+
+        # We only need to check for rewrite if driver is not in WRITE_FULL mode
+        # and if we already stored splits once
+        need_rewrite = (
+            not self.WRITE_FULL
+            and previous_oldest_mutable_timestamp is not None
+        )
+
+        if archive_policy_def.timespan or need_rewrite:
+            existing_keys = self._list_split_keys_for_metric(
+                metric, aggregation, archive_policy_def.granularity)
+
+        # First delete old splits
+        if archive_policy_def.timespan:
             oldest_point_to_keep = ts.last - datetime.timedelta(
                 seconds=archive_policy_def.timespan)
-            self._delete_metric_measures_before(
-                metric, aggregation, archive_policy_def.granularity,
-                oldest_point_to_keep)
+            oldest_key_to_keep = ts.get_split_key(oldest_point_to_keep)
+            oldest_key_to_keep_s = str(oldest_key_to_keep)
+            for key in list(existing_keys):
+                # NOTE(jd) Only delete if the key is strictly inferior to
+                # the timestamp; we don't delete any timeserie split that
+                # contains our timestamp, so we prefer to keep a bit more
+                # than deleting too much
+                if key < oldest_key_to_keep_s:
+                    self._delete_metric_measures(
+                        metric, key, aggregation,
+                        archive_policy_def.granularity)
+                    existing_keys.remove(key)
+        else:
+            oldest_key_to_keep = carbonara.SplitKey(0, 0)
+
+        # Rewrite all read-only splits just for fun (and compression). This
+        # only happens if `previous_oldest_mutable_timestamp' exists, which
+        # means we already wrote some splits at some point – so this is not the
+        # first time we treat this timeserie.
+        if need_rewrite:
+            previous_oldest_mutable_key = str(ts.get_split_key(
+                previous_oldest_mutable_timestamp))
+            oldest_mutable_key = str(ts.get_split_key(
+                oldest_mutable_timestamp))
+
+            if previous_oldest_mutable_key != oldest_mutable_key:
+                for key in existing_keys:
+                    if previous_oldest_mutable_key <= key < oldest_mutable_key:
+                        # NOTE(jd) Rewrite it entirely for fun (and later for
+                        # compression). For that, we just pass None as split.
+                        self._store_timeserie_split(
+                            metric, carbonara.SplitKey(
+                                float(key), archive_policy_def.granularity),
+                            None, aggregation, archive_policy_def,
+                            oldest_mutable_timestamp)
+
+        for key, split in ts.split():
+            if key >= oldest_key_to_keep:
+                self._store_timeserie_split(
+                    metric, key, split, aggregation, archive_policy_def,
+                    oldest_mutable_timestamp)
 
     def add_measures(self, metric, measures):
-        self._store_new_measures(metric, msgpackutils.dumps(
-            list(map(tuple, measures))))
+        measures = list(measures)
+        data = struct.pack(
+            "<" + self._MEASURE_SERIAL_FORMAT * len(measures),
+            *list(
+                itertools.chain(
+                    # NOTE(jd) int(10e8) to avoid rounding errors
+                    *((int(utils.datetime_to_unix(timestamp) * int(10e8)),
+                       value)
+                      for timestamp, value in measures))))
+        self._store_new_measures(metric, data)
 
     @staticmethod
     def _store_new_measures(metric, data):
@@ -226,32 +362,23 @@ class CarbonaraBasedStorage(storage.StorageDriver):
         with self._lock(metric.id)(blocking=sync):
             # If the metric has never been upgraded, we need to delete this
             # here too
-            self._delete_metric_archives(metric)
             self._delete_metric(metric)
 
-    def _delete_metric_measures_before(self, metric, aggregation_method,
-                                       granularity, timestamp):
-        """Delete measures for a metric before a timestamp."""
-        ts = carbonara.AggregatedTimeSerie.get_split_key(
-            timestamp, granularity)
-        for key in self._list_split_keys_for_metric(
-                metric, aggregation_method, granularity):
-            # NOTE(jd) Only delete if the key is strictly inferior to
-            # the timestamp; we don't delete any timeserie split that
-            # contains our timestamp, so we prefer to keep a bit more
-            # than deleting too much
-            if key < ts:
-                self._delete_metric_measures(
-                    metric, key, aggregation_method, granularity)
-
     @staticmethod
     def _delete_metric_measures(metric, timestamp_key,
-                                aggregation, granularity):
+                                aggregation, granularity, version=3):
         raise NotImplementedError
 
-    @staticmethod
-    def _unserialize_measures(data):
-        return msgpackutils.loads(data)
+    _MEASURE_SERIAL_FORMAT = "Qd"
+    _MEASURE_SERIAL_LEN = struct.calcsize(_MEASURE_SERIAL_FORMAT)
+
+    def _unserialize_measures(self, data):
+        nb_measures = len(data) // self._MEASURE_SERIAL_LEN
+        measures = struct.unpack(
+            "<" + self._MEASURE_SERIAL_FORMAT * nb_measures, data)
+        return six.moves.zip(
+            pandas.to_datetime(measures[::2], unit='ns'),
+            itertools.islice(measures, 1, len(measures), 2))
 
     def measures_report(self, details=True):
         metrics, measures, full_details = self._build_report(details)
@@ -263,38 +390,76 @@ class CarbonaraBasedStorage(storage.StorageDriver):
     def _check_for_metric_upgrade(self, metric):
         lock = self._lock(metric.id)
         with lock:
-            for agg_method in metric.archive_policy.aggregation_methods:
+            try:
+                old_unaggregated = self._get_unaggregated_timeserie_and_unserialize_v2(  # noqa
+                    metric)
+            except (storage.MetricDoesNotExist, CorruptionError) as e:
+                # NOTE(jd) This case is not really possible – you can't
+                # have archives with splits and no unaggregated
+                # timeserie…
+                LOG.error(
+                    "Unable to find unaggregated timeserie for "
+                    "metric %s, unable to upgrade data: %s",
+                    metric.id, e)
+                return
+            unaggregated = carbonara.BoundTimeSerie(
+                ts=old_unaggregated.ts,
+                block_size=metric.archive_policy.max_block_size,
+                back_window=metric.archive_policy.back_window)
+            # Upgrade unaggregated timeserie to v3
+            self._store_unaggregated_timeserie(
+                metric, unaggregated.serialize())
+            oldest_mutable_timestamp = (
+                unaggregated.first_block_timestamp()
+            )
+            for agg_method, d in itertools.product(
+                    metric.archive_policy.aggregation_methods,
+                    metric.archive_policy.definition):
                 LOG.debug(
                     "Checking if the metric %s needs migration for %s"
                     % (metric, agg_method))
+
                 try:
-                    data = self._get_metric_archive(metric, agg_method)
+                    all_keys = self._list_split_keys_for_metric(
+                        metric, agg_method, d.granularity, version=2)
                 except storage.MetricDoesNotExist:
                     # Just try the next metric, this one has no measures
                     break
-                except storage.AggregationDoesNotExist:
-                    # This should not happen, but you never know.
-                    LOG.warning(
-                        "Metric %s does not have an archive "
-                        "for aggregation %s, "
-                        "no migration can be done" % (metric, agg_method))
                 else:
                     LOG.info("Migrating metric %s to new format" % metric)
-                    archive = carbonara.TimeSerieArchive.unserialize(data)
-                    for ts in archive.agg_timeseries:
-                        # Store each AggregatedTimeSerie independently
-                        for key, split in ts.split():
-                            self._store_metric_measures(metric, key,
-                                                        ts.aggregation_method,
-                                                        ts.sampling,
-                                                        split.serialize())
-            self._delete_metric_archives(metric)
+                    timeseries = filter(
+                        lambda x: x is not None,
+                        self._map_in_thread(
+                            self._get_measures_and_unserialize_v2,
+                            ((metric, key, agg_method, d.granularity)
+                             for key in all_keys))
+                    )
+                    ts = carbonara.AggregatedTimeSerie.from_timeseries(
+                        sampling=d.granularity,
+                        aggregation_method=agg_method,
+                        timeseries=timeseries, max_size=d.points)
+                    for key, split in ts.split():
+                        self._store_timeserie_split(
+                            metric, key, split,
+                            ts.aggregation_method,
+                            d, oldest_mutable_timestamp)
+                    for key in all_keys:
+                        self._delete_metric_measures(
+                            metric, key, agg_method,
+                            d.granularity, version=None)
+            self._delete_unaggregated_timeserie(metric, version=None)
             LOG.info("Migrated metric %s to new format" % metric)
 
     def upgrade(self, index):
-        self._map_in_thread(
-            self._check_for_metric_upgrade,
-            ((metric,) for metric in index.list_metrics()))
+        marker = None
+        while True:
+            metrics = index.list_metrics(limit=self.UPGRADE_BATCH_SIZE,
+                                         marker=marker)
+            for m in metrics:
+                self._check_for_metric_upgrade(m)
+            if len(metrics) == 0:
+                break
+            marker = metrics[-1].id
 
     def process_new_measures(self, indexer, metrics_to_process, sync=False):
         metrics = indexer.list_metrics(ids=metrics_to_process)
@@ -310,7 +475,7 @@ class CarbonaraBasedStorage(storage.StorageDriver):
                 with self._lock(metric_id)(blocking=sync):
                     self._delete_unprocessed_measures_for_metric_id(metric_id)
             except coordination.LockAcquireFailed:
-                LOG.debug("Cannot acquire lock for metric %s, postponing"
+                LOG.debug("Cannot acquire lock for metric %s, postponing "
                           "unprocessed measures deletion" % metric_id)
         for metric in metrics:
             lock = self._lock(metric.id)
@@ -320,6 +485,7 @@ class CarbonaraBasedStorage(storage.StorageDriver):
             # get back later to it if needed.
             if lock.acquire(blocking=sync):
                 try:
+                    locksw = timeutils.StopWatch().start()
                     LOG.debug("Processing measures for %s" % metric)
                     with self._process_measure_for_metric(metric) as measures:
                         # NOTE(mnaser): The metric could have been handled by
@@ -330,16 +496,13 @@ class CarbonaraBasedStorage(storage.StorageDriver):
                             continue
 
                         measures = sorted(measures, key=operator.itemgetter(0))
+
+                        block_size = metric.archive_policy.max_block_size
                         try:
-                            with timeutils.StopWatch() as sw:
-                                raw_measures = (
-                                    self._get_unaggregated_timeserie(
-                                        metric)
-                                )
-                                LOG.debug(
-                                    "Retrieve unaggregated measures "
-                                    "for %s in %.2fs"
-                                    % (metric.id, sw.elapsed()))
+                            ts = self._get_unaggregated_timeserie_and_unserialize(  # noqa
+                                metric,
+                                block_size=block_size,
+                                back_window=metric.archive_policy.back_window)
                         except storage.MetricDoesNotExist:
                             try:
                                 self._create_metric(metric)
@@ -347,25 +510,21 @@ class CarbonaraBasedStorage(storage.StorageDriver):
                                 # Created in the mean time, do not worry
                                 pass
                             ts = None
-                        else:
-                            try:
-                                ts = carbonara.BoundTimeSerie.unserialize(
-                                    raw_measures)
-                            except ValueError:
-                                ts = None
-                                LOG.error(
-                                    "Data corruption detected for %s "
-                                    "unaggregated timeserie, "
-                                    "recreating an empty one."
-                                    % metric.id)
+                        except CorruptionError as e:
+                            LOG.error(e)
+                            ts = None
 
                         if ts is None:
                             # This is the first time we treat measures for this
                             # metric, or data are corrupted, create a new one
-                            mbs = metric.archive_policy.max_block_size
                             ts = carbonara.BoundTimeSerie(
-                                block_size=mbs,
+                                block_size=block_size,
                                 back_window=metric.archive_policy.back_window)
+                            current_first_block_timestamp = None
+                        else:
+                            current_first_block_timestamp = (
+                                ts.first_block_timestamp()
+                            )
 
                         # NOTE(jd) This is Python where you need such
                         # hack to pass a variable around a closure,
@@ -379,14 +538,16 @@ class CarbonaraBasedStorage(storage.StorageDriver):
                             # affected by new measures for specific granularity
                             tstamp = max(bound_timeserie.first, measures[0][0])
                             computed_points['number'] = len(bound_timeserie)
-                            self._map_in_thread(
-                                self._add_measures,
-                                ((aggregation, d, metric,
-                                  carbonara.TimeSerie(bound_timeserie.ts[
-                                      carbonara.TimeSerie.round_timestamp(
-                                          tstamp, d.granularity * 10e8):]))
-                                 for aggregation in agg_methods
-                                 for d in metric.archive_policy.definition))
+                            for d in metric.archive_policy.definition:
+                                ts = bound_timeserie.group_serie(
+                                    d.granularity, carbonara.round_timestamp(
+                                        tstamp, d.granularity * 10e8))
+                                self._map_in_thread(
+                                    self._add_measures,
+                                    ((aggregation, d, metric, ts,
+                                      current_first_block_timestamp,
+                                      bound_timeserie.first_block_timestamp())
+                                     for aggregation in agg_methods))
 
                         with timeutils.StopWatch() as sw:
                             ts.set_values(
@@ -398,18 +559,29 @@ class CarbonaraBasedStorage(storage.StorageDriver):
                                 len(agg_methods)
                                 * len(metric.archive_policy.definition)
                             )
-                            speed = ((number_of_operations
-                                     * computed_points['number']) / elapsed)
+
+                            if elapsed > 0:
+                                perf = " (%d points/s, %d measures/s)" % (
+                                    ((number_of_operations
+                                      * computed_points['number']) / elapsed),
+                                    ((number_of_operations
+                                      * len(measures)) / elapsed)
+                                )
+                            else:
+                                perf = ""
                             LOG.debug(
                                 "Computed new metric %s with %d new measures "
-                                "in %.2f seconds (%d points/s, %d measures/s)"
-                                % (metric.id, len(measures), elapsed, speed,
-                                   (number_of_operations * len(measures))
-                                   / elapsed))
+                                "in %.2f seconds%s"
+                                % (metric.id, len(measures), elapsed, perf))
 
                         self._store_unaggregated_timeserie(metric,
                                                            ts.serialize())
+
+                    LOG.debug("Metric %s locked during %.2f seconds" %
+                              (metric.id, locksw.elapsed()))
                 except Exception:
+                    LOG.debug("Metric %s locked during %.2f seconds" %
+                              (metric.id, locksw.elapsed()))
                     if sync:
                         raise
                     LOG.error("Error processing new measures", exc_info=True)
@@ -418,11 +590,15 @@ class CarbonaraBasedStorage(storage.StorageDriver):
 
     def get_cross_metric_measures(self, metrics, from_timestamp=None,
                                   to_timestamp=None, aggregation='mean',
+                                  reaggregation=None,
                                   granularity=None,
                                   needed_overlap=100.0):
         super(CarbonaraBasedStorage, self).get_cross_metric_measures(
             metrics, from_timestamp, to_timestamp,
-            aggregation, granularity, needed_overlap)
+            aggregation, reaggregation, granularity, needed_overlap)
+
+        if reaggregation is None:
+            reaggregation = aggregation
 
         if granularity is None:
             granularities = (
@@ -452,7 +628,7 @@ class CarbonaraBasedStorage(storage.StorageDriver):
             return [(timestamp.replace(tzinfo=iso8601.iso8601.UTC), r, v)
                     for timestamp, r, v
                     in carbonara.AggregatedTimeSerie.aggregated(
-                        tss, aggregation, from_timestamp, to_timestamp,
+                        tss, reaggregation, from_timestamp, to_timestamp,
                         needed_overlap)]
         except carbonara.UnAggregableTimeseries as e:
             raise storage.MetricUnaggregatable(metrics, e.reason)
@@ -493,9 +669,40 @@ class CarbonaraBasedStorage(storage.StorageDriver):
 
         return result
 
-    def _map_in_thread(self, method, list_of_args):
+    @staticmethod
+    def _map_no_thread(method, list_of_args):
+        return list(itertools.starmap(method, list_of_args))
+
+    def _map_in_futures_threads(self, method, list_of_args):
         with futures.ThreadPoolExecutor(
                 max_workers=self.aggregation_workers_number) as executor:
             # We use 'list' to iterate all threads here to raise the first
             # exception now, not much choice
             return list(executor.map(lambda args: method(*args), list_of_args))
+
+    @staticmethod
+    def _unserialize_timeserie_v2(data):
+        return carbonara.TimeSerie.from_data(
+            *carbonara.TimeSerie._timestamps_and_values_from_dict(
+                msgpack.loads(data, encoding='utf-8')['values']))
+
+    def _get_unaggregated_timeserie_and_unserialize_v2(self, metric):
+        """Unserialization method for unaggregated v2 timeseries."""
+        data = self._get_unaggregated_timeserie(metric, version=None)
+        try:
+            return self._unserialize_timeserie_v2(data)
+        except ValueError:
+            LOG.error("Data corruption detected for %s ignoring." % metric.id)
+
+    def _get_measures_and_unserialize_v2(self, metric, key,
+                                         aggregation, granularity):
+        """Unserialization method for upgrading v2 objects. Upgrade only."""
+        data = self._get_measures(
+            metric, key, aggregation, granularity, version=None)
+        try:
+            return self._unserialize_timeserie_v2(data)
+        except ValueError:
+            LOG.error("Data corruption detected for %s "
+                      "aggregated `%s' timeserie, granularity `%s' "
+                      "around time `%s', ignoring."
+                      % (metric.id, aggregation, granularity, key))
